/*************************************************************************\
* Copyright (c) 1998-2004 The University of Chicago, as Operator of Argonne
* National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
// CChanObj.cpp : Implementation of CChanObj class

#define CAOBJECT_DLL_SRC

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "CChanObj.h"

#ifdef WIN32
#include <Windows.h>
#endif

// extern "C" wrappers
extern "C" {
    static void exceptionCB(struct exception_handler_args arg);
    static void valueChangedCB(struct event_handler_args args);
    static void connectionChangedCB(
      struct connection_handler_args args);
}

// Static initializations
int CChanObj::caStarted=0;
unsigned long CChanObj::refCount=0;

/***************************** Constructor ****************************/
CChanObj::CChanObj() :
    valueChanged(0),
    userData(NULL),
    Chid(NULL),
    valueChangedEvid(NULL),
    ca_pend_io_time(CA_PEND_IO_TIME),
    connected(0),
    monitored(monitor_off),
    errHandler(errMsg),
    userValueChangedCB(NULL),
    userConnectionChangedCB(NULL)
{
    *value='\0';
    if(!caStarted) taskInit();
    refCount++;
}

/***************************** Destructor *****************************/
CChanObj::~CChanObj()
{
    if(connected || Chid) disconnect();
    refCount--;
    if(refCount <= 0) taskExit();
}

/***************************** taskInit *******************************/
int CChanObj::taskInit(void)
{
  // CA Stuff
    int stat;

    stat=ca_task_initialize();
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::taskInit: "
	  "ca_task_initialize failed:\n%s\n",ca_message(stat));
	return 0;
    }
  
    stat=ca_add_exception_event(::exceptionCB,this);
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::taskInit: "
	  "ca_add_exception_event failed:\n%s\n",ca_message(stat));
	return 0;
    }

    caStarted=1;
    return 1;
}

/***************************** taskExit *******************************/
int CChanObj::taskExit(void)
{
  // CA Stuff
    int stat;

    stat=ca_task_exit();
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::taskExit: "
	  "ca_task_exit failed:\n%s\n",ca_message(stat));
	return 0;
    }
    caStarted=0;
  
    return 1;
}

/***************************** connect ********************************/
int CChanObj::connect(const char *name)
{
    int stat;
    int err=0;

  // Check if connected
    if(connected || Chid) disconnect();
	
  // Search
#if 0
	printf("CChanObj::connect: [%s]\n",name);
#endif
    stat=ca_search(name,&Chid);
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::connect: [%s] "
	  "ca_search failed:\n%s\n",name,ca_message(stat)); err++;
    }
  // (Chid has been allocated if ca_search_and_connect returns ECA_NORMAL)
    stat=ca_pend_io(ca_pend_io_time);
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::connect: [%s] "
	  "ca_pend_io failed:\n%s\n",name,ca_message(stat)); err++;
    }

  // Failure
    if(err) {
	if (Chid != NULL) {
	    stat=ca_clear_channel(Chid);
	    connected=0;
	    value[0]='\0';
	    if(stat != ECA_NORMAL) {
		if(errHandler) (*errHandler)("CChanObj::connect: [%s] "
		  "ca_clear_channel failed:\n%s\n",name,ca_message(stat));
	    }
	    stat=ca_pend_io(ca_pend_io_time);
	    if(stat != ECA_NORMAL) {
		if(errHandler) (*errHandler)("CChanObj::connect: [%s] "
		  "ca_pend_io failed for ca_clear_channel:\n%s\n",name,ca_message(stat));
	    }
	}
	return 0;
    }

  // Success
    connected=1;
    return 1;
}

/***************************** connectCallback ***********************/
int CChanObj::connectCallback(const char *name)
{
    int stat;
    int err=0;

  // Check if connected
    if(connected || Chid) disconnect();
	
  // Search
    stat=ca_search_and_connect(name,&Chid,::connectionChangedCB,
      (void *)this);
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::connectCallback: "
	  "ca_search failed:\n%s\n",ca_message(stat)); err++;
	return 0;
    }
    ca_poll();

  // Success
  // (Chid has been allocated if ca_search_and_connect returns ECA_NORMAL)
    return 1;
}

/***************************** disconnect *****************************/
void CChanObj::disconnect()
{
    int stat;

  // Stop monitoring (Remove value-changed callback)
    stopMonitor();

    if (Chid != NULL) {
#if 0
      // KE: This is not necessary according to Jeff
      // Remove the connection-changed callback
	stat=ca_change_connection_event(Chid,NULL);
	if(stat != ECA_NORMAL) {
	    if(errHandler) (*errHandler)("CChanObj::disconnect: "
	      "ca_change_connection_event failed:\n%s\n",ca_message(stat));
	}
#endif
	
      // Clear the channel
	stat=ca_clear_channel(Chid);
	if(stat != ECA_NORMAL) {
	    if(errHandler) (*errHandler)("CChanObj::disconnect: "
	      "ca_clear_channel failed:\n%s\n",ca_message(stat));
	}
    }
    connected=0;
    Chid=NULL;
    *value='\0';

  // Process requests
    ca_poll();
}

/***************************** get ************************************/
int CChanObj::get(void)
{
    int stat;

  // Check if we think it is connected
    if(!connected) {
	if(errHandler) (*errHandler)("CChanObj::get: "
	  "Channel is not connected.\n");
	return 2;
    }

  // Check if CA thinks it is connected
    if(ca_state(Chid) != cs_conn) {
	if(errHandler)(*errHandler)("CChanObj::get: "
	  "Channel is not connected now.\n");
	return 3;
    }
    
  // Do the get
    stat=ca_array_get(DBR_STRING,1,Chid,value);
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::get: "
	  "ca_array_get:\n%s\n",ca_message(stat));
	return 0;
    }
    stat=ca_pend_io(ca_pend_io_time);
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::get: "
	  "ca_pend_io failed:\n%s\n",ca_message(stat));
	return 0;
    }

  // Success
    return 1;
}

/***************************** getCallback ****************************/
int CChanObj::getCallback(void)
{
    int stat;

  // Check if we think it is connected
    if(!connected) {
	if(errHandler) (*errHandler)("CChanObj::getCallback: "
	  "Channel is not connected.\n");
	return 2;
    }

  // Check if CA thinks it is connected
    if(ca_state(Chid) != cs_conn) {
	if(errHandler) (*errHandler)("CChanObj::getCallback: "
	  "Channel is not connected now.\n");
	return 3;
    }
    
  // Do the get with callback, pass this as the user data
    stat=ca_array_get_callback(DBR_STRING,1,Chid,
      ::valueChangedCB,(void *)this);
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::getCallback: "
	  "ca_array_get_callback:\n%s\n",ca_message(stat));
	return 0;
    }
    ca_poll();

  // Success
    return 1;
}

/***************************** put ************************************/
int CChanObj::put(const char *newval)
{
    int stat;

  // Check if connected
    if(!connected) {
	if(errHandler) (*errHandler)("CChanObj::put: "
	  "Channel is not connected.\n");
	return 2;
    }

  // Check if CA thinks it is connected
    if(ca_state(Chid) != cs_conn) {
	if(errHandler) (*errHandler)("CChanObj::put: "
	  "Channel is not connected now.\n");
	return 3;
    }
    
  // Do the put
    stat=ca_array_put(DBR_STRING,1,Chid,(void *)newval);
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::put: "
	  "ca_array_get:\n%s\n",ca_message(stat));
	return 0;
    }
    stat=ca_pend_io(ca_pend_io_time);
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::put: "
	  "ca_pend_io failed:\n%s\n",ca_message(stat));
	return 0;
    }

  // Success
    return 1;
}

/***************************** putCallback ****************************/
int CChanObj::putCallback(const char *newval)
{
    int stat;

  // Check if connected
    if(!connected) {
	if(errHandler) (*errHandler)("CChanObj::putCallback: "
	  "Channel is not connected.\n");
	return 2;
    }

  // Check if CA thinks it is connected
    if(ca_state(Chid) != cs_conn) {
	if(errHandler) (*errHandler)("CChanObj::putCallback: "
	  "Channel is not connected now.\n");
	return 3;
    }
    
  // Do the put with callback, pass this as the user data
    stat=ca_array_put_callback(DBR_STRING,1,Chid,(void *)newval,
      ::valueChangedCB,(void *)this);
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::putCallback: "
	  "ca_array_get:\n%s\n",ca_message(stat));
	return 0;
    }
    stat=ca_pend_io(ca_pend_io_time);
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::putCallback: "
	  "ca_pend_io failed:\n%s\n",ca_message(stat));
	return 0;
    }

  // Success
    return 1;
}

/***************************** startMonitor ***************************/
int CChanObj::startMonitor()
{
    int stat;

  // Check if connected
    if(!connected) {
	monitored=monitor_req;
	return 2;
    }

  // Check if monitored
    if(monitored == monitor_on) {
	stopMonitor();
    }

    stat=ca_add_array_event(DBR_STRING,1,Chid,::valueChangedCB,
      (void *)this,(float)0,(float)0,(float)0,&valueChangedEvid);
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::startMonitor: "
	  "ca_add_event failed:\n%s\n",ca_message(stat));
	return 0;
    }

  // Poll (Perhaps not necessary)
    ca_poll();

  // Success (assumed if using ca_poll)
    monitored=monitor_on;
    return 1;
}

/***************************** stopMonitor ****************************/
int CChanObj::stopMonitor(void)
{
    int stat;

  // Check if monitored
    if(monitored != monitor_on || !valueChangedEvid) {
	monitored=monitor_off;
	return 1;
    }

    stat=ca_clear_event(valueChangedEvid);
    valueChangedEvid=NULL;
    if(stat != ECA_NORMAL) {
	if(errHandler) (*errHandler)("CChanObj::stopMonitor: "
	  "ca_clear_event failed:\n%s\n",ca_message(stat));
	return 0;
    }

  // Success (assumed)
    monitored=monitor_off;
    return 1;
}

/***************************** errMsg *********************************/
int CChanObj::errMsg(const char *fmt, ...)
{
    static char string[2048];
    va_list vargs;
    int nchars=0;

    va_start(vargs,fmt);
    nchars+=vsprintf(string,fmt,vargs);
    va_end(vargs);

#ifdef WIN32
  //    MessageBox(NULL,string,"CaObject Error",MB_OK|MB_ICONERROR);
#else
    fprintf(stderr,"%s\n",string);
#endif

    return nchars;
}

/***************************** addFDRegistration **********************/
void CChanObj::addFDRegistration(CAFDHANDLER *handler, void *pUser)
{
    int stat;

    stat=ca_add_fd_registration(handler,pUser);
    if(stat != ECA_NORMAL) {
	errMsg("CChanObj::addFDRegistration: "
	  "ca_add_fd_registration failed:\n%s\n",ca_message(stat));
    }
}

/***************************** valueChangedCB *************************/
void CChanObj::valueChangedCB(struct event_handler_args args)
{
    CChanObj *chan=(CChanObj *)args.usr;

  // Copy the value
    if(args.dbr) {
	strcpy(chan->value,(char *)args.dbr);
    } else {
	strcpy(chan->value,"Bad data");
    }
    chan->valueChanged=1;

  // Call the user's value-changed callback
  // The user should set valueChanged=0 (if it wants to)
    if(chan->userValueChangedCB) (*chan->userValueChangedCB)(chan);
}

/***************************** connectionChangedCB ********************/
void CChanObj::connectionChangedCB(struct connection_handler_args args)
{
    chid pCh=args.chid;
    CChanObj *chan=(CChanObj *)ca_puser(pCh);
    pCh=chan->Chid;

  // Branch depending on the state
    switch (ca_state(pCh)) {
    case cs_conn:
	if(!chan->connected) {
	    chan->connected=1;
	  // The following insures the value-changed callback will be called
	    chan->getCallback();
	}
	if(chan->monitored == monitor_req) {
	    chan->startMonitor();
	}
	break;
    case cs_never_conn:
	strcpy(chan->value,"Cannot connect");
	chan->connected=0;
	break;
    case cs_prev_conn:
	strcpy(chan->value,"Lost connection");
	chan->connected=0;
	break;
    case cs_closed:
	strcpy(chan->value,"Invalid");
	chan->connected=0;
	break;
    }

  // Call the users connection callback
  // The value is not guaranteed to be there
    if(chan->userConnectionChangedCB) (*chan->userConnectionChangedCB)(chan);
}

/***************************** exceptionCB ****************************/
void CChanObj::exceptionCB(struct exception_handler_args args)
{
#define MAX_EXCEPTIONS 25    
    static int nexceptions=0;
    static int ended=0;

  // KE: Think about using the user's err handler here as well as below

    if(ended) return;
    if(nexceptions++ > MAX_EXCEPTIONS) {
	ended=1;
	CChanObj::errMsg("CChanObj::exceptionCB: Channel Access Exception:\n"
	  "Too many exceptions [%d]\n"
	  "No more will be handled\n"
	  "Please fix the problem and restart the program",
	  MAX_EXCEPTIONS);
	ca_add_exception_event(NULL,NULL);
	return;
    }

  // Get the time
    char timeStampStr[16];
    long now;
    struct tm *tblock;
    time(&now);
    tblock=localtime(&now);
    strftime(timeStampStr,20,"%b %d %H:%M:%S",tblock);

  // Get the users error handler from args.usr, which is this
    CChanObj *chan=(CChanObj *)args.usr;
    ChanErrorHandler userErrMsg=NULL;
    if(chan) userErrMsg=chan->getErrorHandler();
    if(userErrMsg) userErrMsg("CChanObj::exceptionCB: Channel Access Exception:\n"
      "  Time: %s\n"
      "  Channel Name: %s\n"
      "  Native Type: %s\n"
      "  Native Count: %hu\n"
      "  Access: %s%s\n"
      "  IOC: %s\n"
      "  Message: %s\n"
      "  Context: %s\n"
      "  Requested Type: %s\n"
      "  Requested Count: %ld\n"
      "  Source File: %s\n"
      "  Line number: %u\n",
      timeStampStr,
      args.chid?ca_name(args.chid):"Unavailable",
      args.chid?dbf_type_to_text(ca_field_type(args.chid)):"Unavailable",
      args.chid?ca_element_count(args.chid):0,
      args.chid?(ca_read_access(args.chid)?"R":""):"Unavailable",
      args.chid?(ca_write_access(args.chid)?"W":""):"",
      args.chid?ca_host_name(args.chid):"Unavailable",
      ca_message(args.stat)?ca_message(args.stat):"Unavailable",
      args.ctx?args.ctx:"Unavailable",
      dbf_type_is_valid(args.type)?dbf_type_to_text(args.type):"Invalid",
      args.count,
      args.pFile?args.pFile:"Unavailable",
      args.pFile?args.lineNo:0);
}

// extern "C" wrappers
extern "C" {

    /***************************** valueChangedCB *************************/
static void valueChangedCB(struct event_handler_args args)
{
    CChanObj::valueChangedCB(args);
}

/***************************** connectionChangedCB ********************/
static void connectionChangedCB(struct connection_handler_args args)
{
    CChanObj::connectionChangedCB(args);
}

/***************************** exceptionCB ****************************/
static void exceptionCB(struct exception_handler_args args)
{
    CChanObj::exceptionCB(args);
}

}     // extern "C"
